#!/bin/sh

echo -e "\e[34m=== tools-build.sh"

# Set up the environment to a known state.

SRCDIR=`pwd`/sources
set +h
umask 022
export LC_ALL=POSIX
export CLFAGS="-Os"
export PATH=/sbin:/usr/sbin:/bin:/usr/bin
export HOME=/root
export TERM=linux
unset CC CXX CPP LD_LIBRARY_PATH LD_PRELOAD
export arch=`uname -m`
echo $arch
if [ "$arch" == i686 ]
then
  arch=i386
fi
echo $arch

########## Phase 1-1: Use parent system to build a uclibc cross-compiler

# First, we use the parent system (assumed to be some reasonably sane Linux
# variant) to build a working set of cross-compile build tools.  These tools
# run on the parent system (which could be glibc), but produce binaries linked
# against the uclibc in /tools/lib.  These build tools include a c compiler
# (gcc), a linker and assembler (binutils), and a C library (uclibc) to link
# the resulting binaries against.  Building and using the C library also
# requires a set of kernel ABI headers (matching the linux kernel version to
# be installed on the destination system).

mkdir -p /tools/{bin,etc,tmp,usr} &&

# Unpack linux kernel headers somewhere uclibc can find them.

cd /tools/tmp &&
tar xvjf "$SRCDIR/buildtools/linux-libc-headers.tar.bz2" &&
cd linux-libc-headers*/include &&
mv asm-$arch ../../asm &&
mv linux sound ../.. &&
cd ../.. &&
rm -rf linux-libc-headers*

if [ $? -ne 0 ]; then exit 1; fi

echo === Build uclibc

cd /tools/tmp &&
tar xvjf "$SRCDIR"/system/uClibc.tar.bz2 &&
cd uClibc* &&
cat "$SRCDIR"/system/uClibc*patch | patch -p1 &&
cat "$SRCDIR"/system/uClibc-config > .config &&
# fix_includes gets deeply confused by pre-prepared kernel headers, so
# do its job manually.
echo > extra/scripts/fix_includes.sh &&
mv ../asm ../linux ../sound include &&
make &&
# Install's a bit funky, but this seems to work.
make RUNTIME_PREFIX=/tools/ DEVEL_PREFIX=/tools/ install_runtime install_dev &&
# make clean breaks if asm, linux, and scsi aren't symlinks...
#mv include/asm include/linux include/scsi . &&
#make clean &&
#mv asm linux scsi include

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Build binutils that links against /tools"

cd /tools/tmp &&
tar xvjf "$SRCDIR"/buildtools/binutils.tar.bz2 &&
mv binutils-* binutils-source &&
sed -i -e 's@^NATIVE_LIB_DIRS.*@NATIVE_LIB_DIRS=/tools/lib@' binutils-source/ld/configure.tgt &&
mkdir binutils-build &&
cd binutils-build &&
../binutils-source/configure --prefix=/tools --disable-nls &&
make configure-host &&
make &&
make install &&
cd .. &&
rm -rf binutils-*

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Build gcc that links against /tools"
# Note: we need to modify gcc source a LOT, both to point various paths at
# the /tools directory, and to point them at uclibc instead of glibc.
# In theory we should be able to tell ./configure "--prefix=/tools" and
# "--arch=$(`uname -m`)-pc-linux-uclibc", and have everything else just work.
# Unfortunately, this doesn't actually work...

cd /tools/tmp &&
tar xvjf "$SRCDIR"/buildtools/gcc-core.tar.bz2 &&
mv gcc-* gcc-source
cd gcc-source &&
# Do NOT run fixincludes (Replaces no_fixincludes patch from LFS)
sed -i -e 's@\./fixinc\.sh@-c true@' -e 's@.*include/README@	echo@' \
	gcc/Makefile.in &&
## Fix pointlessly hardwired paths.
#sed -i 's@\(standard_startfile_prefix_[12].*\)"/@&tools/@g' gcc/gcc.c &&
# Remove pointlessly hardwired paths.
sed -i -e 's@(\*standard_startfile_prefix_[12])@(0)@' gcc/gcc.c &&
# Edit other pointlessly hardwired paths
sed -r -i -e 's@.*"/usr/lib".*@@' -e 's@.*"/usr/local/lib".*@@' -e 's@"/lib@"/tools/lib@g' -e 's@"/usr@"/tools/usr@g' gcc/collect2.c &&
# Adjust config entries from which spec file will be generated
for i in `find gcc/config -name "*.h"`
do
  sed -i -r -e 's /lib(64)?/ld(64)?(-linux)?(-x86-)?(-ia)?(64)?.so.[12] /tools/lib/ld-uClibc.so.0 ' -e 's P,/usr/lib P,/tools/lib ' -e 's@-rpath /lib@-rpath /tools/lib@' "$i"
  if [ $? -ne 0 ]; then exit 1; fi
done &&
# Remove /usr/include from the end of the include search path.
# (Also part of specs-2 patch from LFS)
cat >> gcc/config/linux.h << EOF &&
#undef STANDARD_INCLUDE_DIR
#define STANDARD_INCLUDE_DIR 0
EOF
# Chop out a stack unwinding symbol uClibc doesn't provide.
#sed -i -e 's/(dl_iterate_phdr.*< 0)/(0)/' gcc/unwind-dw2-fde-glibc.c &&
# We don't need this, and it would break for the second build, so just yank it.
rm -rf libmudflap &&
# We've adjusted the source, now build it.
cd .. &&
mkdir gcc-build &&
cd gcc-build &&
../gcc-source/configure --prefix=/tools --with-local-prefix=/tools --disable-nls --enable-shared --enable-languages=c &&
# GCC's normal build process (make bootstrap) actually builds gcc three times:
# first with the parent system's compiler, then rebuilds it using the compiler
# it built in the first pass, and finally builds a third instance using the
# second one to make sure they produce identical binaries.  (This level of
# paranoia comes from cross-compiling gcc from Solaris or HP-UX using the
# system's wonky native compilers.)
#
# The "quickstrap" target just builds the first pass and installs it.
# We're building from a Linux system, and we're going to recompile it with
# itself anyway in phase 1-2.
make quickstrap &&
make install &&
ln -sf gcc /tools/bin/cc &&
cd .. &&
rm -rf gcc-*

if [ $? -ne 0 ]; then exit 1; fi

# The phase 1 toolset is complete and installed.  All we have to do to use
# it is put /tools/bin at the start of the path, so the binaries it contains
# are found before the ones in the host system.
export PATH=/tools/bin:"$PATH"

# Quick test to make sure the toolset is creating binaries linked against
# /tools/lib and uclibc.
echo 'main(){}' > dummy.c &&
gcc dummy.c &&
xtest=`readelf -l a.out | grep ': /tools/lib'` &&
if [ -z "$xtest" ]
then
  echo "error: Toolchain isn't linking against /tools directory"
  exit 1
fi
# Now make sure there's nothing _but_ uclibc getting linked in.
xtest=`ldd a.out | grep -v "libc.so.0" | grep -v "ld-uClibc.so.0"`
if [ -z $xtest ]
then
  echo "Toolchain diagnostics check out."
else
  echo "error: Toolchain is not creating uclibc-only dynamically linked executables."
  exit 1
fi
unset xtest
rm -f dummy.c a.out

# Phase 2: Add utilities.
#
# Everything we build with the new toolchain will depend on /tools/lib, so
# we can drop a nice self-contained /tools directory into an otherwise empty
# directory, chroot into the empty directory to build stuff, and then cleanly
# remove /tools when we're done.
#
# A compiler alone isn't enough to build stuff with.  In theory adding busybox
# and make should suffice.  (In practice, busybox isn't quite there yet, so
# we need bash too.)
#
# Since we still have the uClibc source lying around, let's start with that.
# We don't have to reinstall uClibc (even though this version was built by
# the parent system's compiler, it'll get rebuilt anyway for the final
# install.)  But we do need versions of ldd and such that link against /tools.

echo === Build uclibc utilities. &&
cd /tools/tmp/uClibc* &&
# We ran the library build earlier, so the includes directory should have
# the correct uClibc headers in it.
make RUNTIME_PREFIX=/tools/ DEVEL_PREFIX=/tools/ install_utils &&
cd .. &&
rm -rf uClibc*

if [ $? -ne 0 ]; then exit 1; fi

echo === busybox

# Alas, this is different enough from build-busybox.sh to not easily be merged.

cd /tools/tmp &&
tar xvjf "$SRCDIR"/system/busybox.tar.bz2 &&
cd busybox* &&
cat "$SRCDIR"/system/busybox-config-firmware > .config &&
yes "" | make oldconfig &&
make &&
cp busybox /tools/bin/ &&
rm /tools/bin/strings && # Use busybox version, not Binutils version.
(for i in `sed 's ^.*/  ' busybox.links`; do ln -s busybox "/tools/bin/$i"; done) &&
cd .. &&
rm -rf busybox*

if [ $? -ne 0 ]; then exit 1; fi

echo === make

cd /tools/tmp &&
tar xvjf "$SRCDIR"/buildtools/make.tar.bz2 &&
cd make-* &&
./configure --prefix=/tools &&
make &&
make install &&
cd .. &&
rm -rf make-*

if [ $? -ne 0 ]; then exit 1; fi

# Someday, the three busybox shells should be unified and extended into a
# workable bash replacement.  Until then, we supplement busybox with bash.

echo === bash

cd /tools/tmp &&
tar xvzf "$SRCDIR"/temp/bash.tar.gz &&
cd bash* &&
cat "$SRCDIR"/temp/bash*patch | patch -p1 &&
# Fix bash's default paths (mostly because "/usr/gnu/bin" is just stupid).
sed -i 's@".*/usr/bin.*"@"/tools/bin"@' config-top.h &&
./configure --prefix=/tools &&
make &&
make install &&
cd .. &&
rm -rf bash* &&
ln -s bash /tools/bin/sh

if [ $? -ne 0 ]; then exit 1; fi

# Phase 3: Rebuild binutils and gcc
#
# Problem 1: The gcc and binutils binaries we built earlier are linked against
# the parent system.  Although they generate stuff that can work when /tools
# stands alone in a chroot environment, they themselves can't run from there.
#
# Problem 2: When we use /tools to build a new environment, we want the new
# binaries we build to work without /tools.
#
# Building a new compiler with the current compiler will fix problem 1,
# and we can configure the new compiler to fix problem 2.

echo === rebuild binutils dynamically linked against uclibc.

cd /tools/tmp &&
tar xvjf "$SRCDIR"/buildtools/binutils.tar.bz2 &&
mv binutils-* binutils-source &&
mkdir binutils-build &&
cd binutils-build &&
../binutils-source/configure --prefix=/tools --enable-shared && #--with-lib-path=/tools/lib &&
make NATIVE_LIB_DIRS=/lib &&
make install &&
cd .. &&
rm -rf binutils-*

if [ $? -ne 0 ]; then exit 1; fi

echo === Rebuild gcc with /tools, dynamically linked against uclibc.

cd /tools/tmp &&
tar xvjf "$SRCDIR"/buildtools/gcc-core.tar.bz2 &&
mv gcc-* gcc-source &&
cd gcc-source &&
# Do NOT run fixincludes (Replaces no_fixincludes patch from LFS)
sed -i -e 's@\./fixinc\.sh@-c true@' -e 's@.*include/README@	echo@' \
	gcc/Makefile.in &&
# Remove pointlessly hardwired paths.
sed -i -e 's@(\*standard_startfile_prefix_[12])@(0)@' gcc/gcc.c &&
# Libmudflap will try to build using the new compiler before it's even
# installed, meaning # it'll complain there's no /lib/uClibc* yet.  Well duh.
rm -rf libmudflap &&
# Don't install redundant libiberty
sed -i "s/install:.*install-subdir/install: install-subdir/" libiberty/Makefile.in &&
for i in `find gcc/config -name "*.h"`
do
  sed -i -r -e 's /lib(64)?/ld(64)?(-linux)?(-x86-)?(-ia)?(64)?.so.[12] /lib/ld-uClibc.so.0 ' "$i"
  if [ $? -ne 0 ]; then exit 1; fi
done &&
mkdir ../gcc-build &&
cd ../gcc-build &&
../gcc-source/configure --prefix=/tools --with-local-prefix=/tools --disable-nls --enable-shared --enable-languages=c &&
#--enable-threads=posix --enable-__cxa_atexit --enable-clocale=gnu &&
# Once again, build just once rather than three times.  It's already being
# built with itself, and we'll build it again later in the final system.
make quickstrap &&
make install &&
cd .. &&
rm -rf gcc-*

if [ $? -ne 0 ]; then exit 1; fi

echo === strip debugging symbols.

# This can fail, so don't check error returns.
rm -rf /tools/{doc,info,man,share/locale}
find /tools/ -type f | xargs strip --strip-debug

# Let parent script know we completed successfully.
touch /tools/finished
